@@ -61,6 +61,11 @@ class Homee:
     async def get_access_token(self):
         """Asynchronously attempts to get an access token from the homee host using username and password."""
 
+        
+        # Check if current token is still valid
+        if self.token != None and self.expires > datetime.now().timestamp():
+            return self.token
+        
         client = aiohttp.ClientSession()
         auth = BasicAuth(
             self.user, hashlib.sha512(self.password.encode("utf-8")).hexdigest()
@@ -75,21 +80,11 @@ class Homee:
             "device_app": DeviceApp.HOMEE,
         }
 
-        if self.token != None and self.expires > datetime.now().timestamp():
-            return self.token
-
-        if self.retries > 0:
-            self.reconnect()
-
-        self.retries += 1
-        if self.retries > self.maxRetries:
-            raise Exception("max retries reached")
-
         try:
             req = await client.post(
                 url, auth=auth, data=data, headers=headers, timeout=5
             )
-        except asyncio.TimeoutError as e:
+        except Exception as e:
             await client.close()
             raise e
 
@@ -118,8 +113,15 @@ class Homee:
         try:
             await self.get_access_token()
         except:
-            raise AuthenticationFailedException
-
+            # Attempt to reconnect if there was a timeout during authentication
+            if self.retries < self.maxRetries:
+                self.retries += 1
+                await self.reconnect()
+                return
+            else:
+                await self.on_max_retries()
+                raise AuthenticationFailedException
+        
         await self.open_ws()
 
     def start(self):
@@ -132,11 +134,10 @@ class Homee:
         self._log("Opening websocket...")
 
         if self.retries > 0:
-            self.on_reconnect()
-
-        self.retries += 1
+            await self.on_reconnect()
+        
         if self.retries > self.maxRetries:
-            self.on_max_retries()
+            await self.on_max_retries()
             return
 
         try:
@@ -175,12 +176,14 @@ class Homee:
 
                     except websockets.exceptions.ConnectionClosed as e:
                         self.connected = False
+                        ws.abort_pings()
                         await self.on_disconnected()
         except Exception as e:
             # TODO retry logic
             await self._ws_on_error(e)
             # raise e
-
+        
+        self.retries += 1
         await self._ws_on_close()
 
     async def _ws_receive_handler(self, ws: websockets.WebSocketClientProtocol):
@@ -206,7 +209,7 @@ class Homee:
         if self.pingInterval <= 0:
             return
 
-        while self.connected and not self.shouldClose:
+        while self.connected and not self.shouldClose and ws.open:
             await ws.ping()
             self._log("PING!")     
             await asyncio.sleep(self.pingInterval)
@@ -217,9 +220,10 @@ class Homee:
         self._log("Connection to websocket successfull")
 
         self.connected = True
-        self.retries = 1
-
+        
         await self.on_connected()
+        self.retries = 0
+
         await self.send("GET:all")
 
     async def _ws_on_message(self, msg: str):
@@ -258,7 +262,7 @@ class Homee:
         self._log(f"Attempting to reconnect in {self.reconnectInterval * self.retries} seconds...")
         
         await asyncio.sleep(self.reconnectInterval * self.retries)
-        await self.open_ws()
+        await self.run()
 
     def disconnect(self):
         """Disconnect from homee by closing the websocket connection."""
@@ -425,10 +429,9 @@ class Homee:
         return self._disconnected_event.wait()
 
     def on_reconnect(self):
-        """TODO"""
-        pass
+        """Called right before a reconnection attempt is started."""
 
-    def on_max_retries(self):
+    async def on_max_retries(self):
         """Called if the maximum amount of retries was reached."""
 
     async def on_connected(self):
